/*
 * forklift scheduling problem from Graeme Gange
 *
 */

#include "CFGC-Propagator.h"
#include <time.h>
#include <math.h>
#include <iomanip>

#include <gecode/driver.hh>
#include <gecode/minimodel.hh>

using namespace Gecode;

struct DFA_Transition
{
	int from_state;
	int to_state;
	int arc;

	DFA_Transition(int t_from_state, int t_to_state, int t_arc)
	{
		from_state = t_from_state;
		to_state = t_to_state;
		arc = t_arc;
	}
};

struct DFA_Transition_List
{
	DFA_Transition* cur_transition;
	DFA_Transition_List* next;

	DFA_Transition_List()
	{
		cur_transition = NULL;
		next = NULL;
	}
};

DFA_Transition_List** dfa_tran_graph_list_head;
DFA_Transition_List** dfa_tran_graph_list_end;
int* dfa_tran_count;
int* start_state;
set<int>* accept_states;

string cfg_file_name;
string benchmark_file_name;
int val_dom_min;
int val_dom_max;
int num_station;
int num_item;
int sched_horizon;
int* source;
int* destination;

bool input_file_error;

bool gecode_find_solution;
map<string,int> string_to_terminal;
map<int,string> terminal_name;
ParseGraph* parse_graph_pr;
DFA::Transition** tran_graph;

bool readBenchmark(string benchmark_file_name, int* source, int* destination);

class Forklift_Scheduling : public MinimizeScript {

	IntVarArray dec_variables;
	IntVarArray c_act;
	IntVar c_total;

public:

	Forklift_Scheduling(const Options& opt)
	: dec_variables(*this, sched_horizon, val_dom_min, val_dom_max),
	  c_act(*this, sched_horizon, 0, 3){

		gecode_find_solution = false;

		parse_graph_pr = new ParseGraph(cfg_file_name, sched_horizon,
				string_to_terminal, terminal_name, val_dom_max+1);

		input_file_error = parse_graph_pr->input_file_error;

		if(input_file_error)
		{
			cout << "Error exists in the grammar file!" << endl;
			return;
		}

		//  compute the scheduling cost

		IntArgs c_array(1+val_dom_max);
		c_array[0] = 0;
		for(int i=1; i<=num_station; i++)
		{
			c_array[i] = 3;
		}
		for(int i=num_station+1; i<=val_dom_max; i++)
		{
			c_array[i] = 1;
		}

		IntSharedArray cs_array(c_array);

		for(int i=0; i<sched_horizon; i++)
		{
			element(*this, cs_array, dec_variables[i], c_act[i]);
		}

/*		for(int i=0; i<sched_horizon; i++)
		{
			BoolVar b1,b2;
			b1 = expr(*this, dec_variables[i]==0);
			b2 = expr(*this, c_act[i]==0);
			rel(*this, b1, BOT_IMP, b2, 1);

			BoolVar b3,b4;
			b3 = expr(*this, (dec_variables[i]>0)&&(dec_variables[i]<=num_station));
			b4 = expr(*this, c_act[i]==3);
			rel(*this, b3, BOT_IMP, b4, 1);

			BoolVar b5,b6;
			b5 = expr(*this, dec_variables[i]>num_station);
			b6 = expr(*this, c_act[i]==1);
			rel(*this, b5, BOT_IMP, b6, 1);
		}*/

		c_total = expr(*this, sum(c_act));


		cfgc(*this, dec_variables, parse_graph_pr, 1);

		// TODO: add the regular constraints here

		tran_graph = new DFA::Transition*[num_item];

		for(int i=0; i<num_item; i++)
		{
			int transition_num = dfa_tran_count[i]+1;
			tran_graph[i] = new DFA::Transition[transition_num];

		    int j = 0;
		    DFA_Transition_List* dfa_transition_list_cur =
		    		dfa_tran_graph_list_head[i];

		    while(dfa_transition_list_cur!=NULL)
		    {
		    	DFA_Transition* cur_transition =
		    			dfa_transition_list_cur->cur_transition;
		    	tran_graph[i][j].i_state = cur_transition->from_state;
		    	tran_graph[i][j].o_state = cur_transition->to_state;
		    	tran_graph[i][j].symbol = cur_transition->arc;

		    	j++;
		    	dfa_transition_list_cur = dfa_transition_list_cur->next;
  		    }

	    	tran_graph[i][j].i_state = -1;
	    	tran_graph[i][j].o_state = 0;
	    	tran_graph[i][j].symbol = 0;

		    int i_start_state = start_state[i];
		    int t_size = accept_states[i].size();
		    int* i_final_state = new int[t_size+1];

		    set<int>::iterator set_int_it;
		    j = 0;
			for(set_int_it=accept_states[i].begin();
					set_int_it!=accept_states[i].end(); set_int_it++)
			{
				i_final_state[j] = *set_int_it;
				j++;
			}
			i_final_state[j] = -1;

		    DFA myDFA(i_start_state, tran_graph[i], i_final_state);
		    extensional(*this, dec_variables, myDFA);

		    delete[] i_final_state;
		}

		branch(*this, dec_variables, INT_VAR_SIZE_MIN, INT_VAL_MIN);
	}

	~Forklift_Scheduling()
	{
	}


	Forklift_Scheduling(bool share, Forklift_Scheduling& s)
    : MinimizeScript(share,s)
	{
		dec_variables.update(*this, share, s.dec_variables);
		c_act.update(*this, share, s.c_act);
		c_total.update(*this, share, s.c_total);
	}

	virtual IntVar cost(void) const
	{
	    return c_total;
	}

	virtual Space* copy(bool share)
	{
		return new Forklift_Scheduling(share,*this);
	}

	virtual void print(std::ostream& os) const
	{
		if(!input_file_error)
		{
			os << std::endl << "Find the following solution with a cost of "
					<< c_total.val() << " :" << std::endl;

			os << std::endl << "The forklift does the following actions:" << std::endl;

			gecode_find_solution = true;

			for (int i=0; i<sched_horizon; i++)
			{
                int act_ind = dec_variables[i].val();
                os << terminal_name[act_ind] << std::endl;
			}

			os << std::endl;
		}
	}
};

// --- main for script run with parameters from argv --- //
int main(int argc, char* argv[]) {

	/*if(argc!=4)
	{
		cout << "Wrong parameters for running this program!!!" << endl;
		return 0;
	}

	num_station = atoi(argv[1]);
	num_item = atoi(argv[2]);
	sched_horizon = atoi(argv[3]);*/


	num_station = 3;
	num_item = 5;
	sched_horizon = 18;

	clock_t start_time = clock();

	cout << endl
		 << "Now solving the instance: "
		 << num_station << " stations, "
		 << num_item << " items, and "
		 << sched_horizon << " scheduling horizon!" << endl;

	benchmark_file_name = "/home/jun/workspace/ForkliftScheduling-IMP/benchmark/";
	stringstream str_stream;
	str_stream << num_station << "_" << num_item << ".wrh";
	benchmark_file_name.append(str_stream.str());

	source = new int[num_item];
	destination = new int[num_item];

	if(!readBenchmark(benchmark_file_name, source, destination))
	{
		cout << "Error exists in the benchmark file!" << endl;
		return 0;
	}

	val_dom_min = 0;
	val_dom_max = num_station + 2*num_item;

	string terminal_str;

	terminal_str = "idle";
	string_to_terminal[terminal_str] = 0;
	terminal_name[0] = terminal_str;

	for(int i=1; i<=num_station; i++)
	{
		str_stream.str("");
		str_stream << "move to station " << i-1;
		terminal_str = str_stream.str();
		string_to_terminal[terminal_str] = i;
		terminal_name[i] = terminal_str;
	}

	int item_id = 0;
	for(int i=num_station+1; i<=num_station+num_item; i++)
	{
		str_stream.str("");
		str_stream << "load item " << item_id;
		terminal_str = str_stream.str();
		string_to_terminal[terminal_str] = i;
		terminal_name[i] = terminal_str;
		item_id++;
	}

	item_id = 0;
	for(int i=num_station+num_item+1; i<=val_dom_max; i++)
	{
		str_stream.str("");
		str_stream << "unload item " << item_id;
		terminal_str = str_stream.str();
		string_to_terminal[terminal_str] = i;
		terminal_name[i] = terminal_str;
		item_id++;
	}

	// write the grammar file

	cfg_file_name = "/home/jun/workspace/ForkliftScheduling-IMP/results/CFG/";
	str_stream.str("");
	str_stream << num_station << "_" << num_item << ".cfg";
	cfg_file_name.append(str_stream.str());

	ofstream cfgFile;
	cfgFile.open(cfg_file_name.c_str());

	cfgFile << "cfg S :=  W | W I ;" << endl ;
	cfgFile << "cfg I := \"idle\" I | \"idle\" ;" << endl;
	cfgFile << "cfg W := W W ;" << endl;

	for(int i=0; i<num_station; i++)
	{
		cfgFile << "cfg W := \"move to station " << i << "\" ;" << endl;
	}

	for(int i=0; i<num_item; i++)
	{
		cfgFile << "cfg W := \"load item " << i << "\" W \"unload item " << i
				<< "\" ;" << endl;
	}

	cfgFile.close();

	// build the automata for each item

	dfa_tran_graph_list_head = new DFA_Transition_List*[num_item];
	dfa_tran_graph_list_end = new DFA_Transition_List*[num_item];
	dfa_tran_count = new int[num_item];
	start_state = new int[num_item];
	accept_states = new set<int>[num_item];

	int** map_to_state_id = new int*[num_station];
	for(int s=0; s<num_station; s++)
	{
		map_to_state_id[s] = new int[3];
	}
	int* state_id_to_station = new int[num_station*3];
	int* state_id_to_status = new int[num_station*3];

	int t_id = 0;
	for(int s=0; s<num_station; s++)
	{
		for(int it=0; it<3; it++)
		{
			map_to_state_id[s][it] = t_id;
			state_id_to_station[t_id] = s;
			state_id_to_status[t_id] = it;

			t_id++;
		}
	}

	for(int i=0; i<num_item; i++)
	{
		dfa_tran_count[i] = 2;
		dfa_tran_graph_list_head[i] = new DFA_Transition_List();
		dfa_tran_graph_list_end[i] = dfa_tran_graph_list_head[i];

		int i_source = source[i];
		int i_destination = destination[i];
		start_state[i] = map_to_state_id[i_source][0];

		accept_states[i].clear();
		int t_accept_state1 = map_to_state_id[i_destination][0];
		int t_accept_state2 = map_to_state_id[i_destination][1];
		accept_states[i].insert(t_accept_state1);
		accept_states[i].insert(t_accept_state2);

		dfa_tran_graph_list_end[i]->cur_transition =
				new DFA_Transition(t_accept_state1, t_accept_state1, 0);

		dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
		dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
		dfa_tran_graph_list_end[i]->cur_transition =
				new DFA_Transition(t_accept_state2, t_accept_state2, 0);

		for(int k_station=0; k_station<num_station; k_station++)
		{
			int from_state = map_to_state_id[k_station][0];

			for(int j_station=0; j_station<num_station; j_station++)
			{
				dfa_tran_count[i]++;

				int arc = j_station+1;
				int to_state;
				if(k_station==j_station)  to_state = map_to_state_id[k_station][1];
				else to_state = from_state;

				dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
				dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
				dfa_tran_graph_list_end[i]->cur_transition =
						new DFA_Transition(from_state, to_state, arc);
			}

			for(int j_item=0; j_item<num_item; j_item++)
			{
				if(j_item!=i)
				{
					dfa_tran_count[i]++;

					int arc = num_station+j_item+1;
					int to_state = from_state;

					dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
					dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
					dfa_tran_graph_list_end[i]->cur_transition =
							new DFA_Transition(from_state, to_state, arc);
				}
			}

			for(int j_item=0; j_item<num_item; j_item++)
			{
				if(j_item!=i)
				{
					dfa_tran_count[i]++;

					int arc = num_station+num_item+j_item+1;
					int to_state = from_state;

					dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
					dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
					dfa_tran_graph_list_end[i]->cur_transition =
							new DFA_Transition(from_state, to_state, arc);
				}
			}

			//  the second line

			from_state = map_to_state_id[k_station][1];

			for(int j_station=0; j_station<num_station; j_station++)
			{
				if(k_station!=j_station)
				{
					dfa_tran_count[i]++;

					int arc = j_station+1;
					int to_state = map_to_state_id[k_station][0];

					dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
					dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
					dfa_tran_graph_list_end[i]->cur_transition =
							new DFA_Transition(from_state, to_state, arc);
				}
			}

			for(int j_item=0; j_item<num_item; j_item++)
			{
				dfa_tran_count[i]++;

				int arc = num_station+j_item+1;
				int to_state;

				if(j_item!=i)  to_state = from_state;
				else  to_state = map_to_state_id[k_station][2];

				dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
				dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
				dfa_tran_graph_list_end[i]->cur_transition =
						new DFA_Transition(from_state, to_state, arc);
			}

			for(int j_item=0; j_item<num_item; j_item++)
			{
				if(j_item!=i)
				{
					dfa_tran_count[i]++;

					int arc = num_station+num_item+j_item+1;
					int to_state = from_state;

					dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
					dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
					dfa_tran_graph_list_end[i]->cur_transition =
							new DFA_Transition(from_state, to_state, arc);
				}
			}

			//  the third line

			from_state = map_to_state_id[k_station][2];

			for(int j_station=0; j_station<num_station; j_station++)
			{
				if(k_station!=j_station)
				{
					dfa_tran_count[i]++;

					int arc = j_station+1;
					int to_state = map_to_state_id[j_station][2];

					dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
					dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
					dfa_tran_graph_list_end[i]->cur_transition =
							new DFA_Transition(from_state, to_state, arc);
				}
			}

			for(int j_item=0; j_item<num_item; j_item++)
			{
				if(j_item!=i)
				{
					dfa_tran_count[i]++;

					int arc = num_station+j_item+1;
					int to_state = from_state;

					dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
					dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
					dfa_tran_graph_list_end[i]->cur_transition =
							new DFA_Transition(from_state, to_state, arc);
				}
			}

			for(int j_item=0; j_item<num_item; j_item++)
			{
				dfa_tran_count[i]++;

				int arc = num_station+num_item+j_item+1;
				int to_state;

				if(j_item!=i)  to_state = from_state;
				else  to_state = map_to_state_id[k_station][1];

				dfa_tran_graph_list_end[i]->next = new DFA_Transition_List();
				dfa_tran_graph_list_end[i] = dfa_tran_graph_list_end[i]->next;
				dfa_tran_graph_list_end[i]->cur_transition =
						new DFA_Transition(from_state, to_state, arc);
			}
		}
	}

	clock_t end_time1 = clock();

	ofstream resFile;
	resFile.open("/home/jun/workspace/ForkliftScheduling-IMP/results/cfgc_imp.res");

	Options opt("Forklift-Scheduling");
	opt.solutions(0);
	opt.time(60000*60*5);
	opt.parse(argc,argv);

	MinimizeScript::run<Forklift_Scheduling,BAB,Options>(opt);

	clock_t end_time2 = clock();

	if(gecode_find_solution)
	{
		resFile << 1 << endl;
	}
	else resFile << 0 << endl;

	resFile << (double)(end_time2-start_time)*1000.0f/CLOCKS_PER_SEC << endl;

	cout << endl
	     << "The instance is now solved in "
	     << (end_time2-start_time)*1000.0f/CLOCKS_PER_SEC
	     << ", with " << (end_time2-end_time1)*1000.0f/CLOCKS_PER_SEC
	     << " for constraint solving" << endl;

	delete parse_graph_pr;

	delete[] source;
	delete[] destination;

	delete[] dfa_tran_count;
	delete[] start_state;
	delete[] accept_states;

	for(int i=0; i<num_item; i++)
	{
		DFA_Transition_List* cur_list = dfa_tran_graph_list_head[i];
		while(cur_list!=NULL)
		{
			if(cur_list->cur_transition==NULL)
			{
				delete cur_list;
				break;
			}

			delete cur_list->cur_transition;

			DFA_Transition_List* temp_pr = cur_list;
			cur_list = cur_list->next;
			delete temp_pr;
		}

		delete[] tran_graph[i];
	}
	delete[] tran_graph;

	delete[] dfa_tran_graph_list_head;
	delete[] dfa_tran_graph_list_end;

	for(int s=0; s<num_station; s++)
	{
		delete[] map_to_state_id[s];
	}
	delete[] map_to_state_id;

	delete[] state_id_to_station;
	delete[] state_id_to_status;

	return 1;
}

bool readBenchmark(string benchmark_file_name, int* source, int* destination)
{
	int num_station, num_item;

	ifstream inFile(benchmark_file_name.c_str());

	if (inFile.is_open())
	{
		inFile >> num_station;
		inFile >> num_item;

		for(int i=0; i<num_item; i++)
		{
			inFile >> source[i];
			inFile >> destination[i];
		}

		return true;
	}
	else return false;
}



